/* strings -- print the strings of printable characters in files
   Copyright (C) 1993, 94 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Written by Richard Stallman <rms@gnu.ai.mit.edu>
   and David MacKenzie <djm@gnu.ai.mit.edu>.

   CC65 / ACE16 (Commodore 64/128 o.s.) version
   by Josef Soucek <josef.soucek@spt.cz> 6/9/1998 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>     /* isprint() */
#include <string.h>
/* #include <ace.h> */

/* Definice na zacatku */

typedef unsigned file_ptr;

#define true 1
#define false 0
#define isgraphic(c) (isascii_my (c))
          
/* Prototypy */

int strings_file (char *file);
int print_strings (char *filename, FILE *stream, file_ptr address, int stop_point);
int usage (FILE *stream, int status);
int isascii_my (int c);

/* Externi promene */

/* Radix for printing addresses (must be 8, 10 or 16).  */
static int address_radix;

/* true means print address within file for each string.  */
static int print_addresses;

/* true means print filename for each string.  */
static int print_filenames;

/* minimalni mnozstvi znaku povazovene za string */
static int string_min;

char *program_name;

/* MAIN */

void main (int argc, char **argv)
{
  int argument;
  argument = 1;
  program_name = argv[0];
  
  string_min = 4;            /* minimalni velikost stringu */
  print_addresses = false;
/*  address_radix = 16;      */
  print_filenames = false;
  
/*  printf("argc=%d\n",argc);  */

    if ((!strcmp(argv[argument],"-t")) && (argc >= argument+2))
      {
/*      printf("detekovan -t\n"); */
      argument++;
      print_addresses=true;
      switch ((int) *argv[argument++]) 
        {
        case 'o':
          address_radix = 8;
          break;
      
        case 'd':
          address_radix = 10;
          break;

        case 'x':
          address_radix = 16;
          break;

        default:
          usage (stderr, EXIT_FAILURE);
        }
      }
    if ((!strcmp(argv[argument],"-n")) && (argc >= argument+2))
      {
/*      printf("detekovan -n\n"); */
      argument++;
      string_min = atoi(argv[argument++]);
      if (string_min < 1)
        {
        fprintf(stderr,"%s: invalid number %s\n", program_name, argv[argument-1]);
        exit (EXIT_FAILURE);
        }
      }
/*    printf("argc=%d argument=%d\n",argc,argument); */
    if (argc == argument) exit(strings_file(NULL));
    if (argc != argument+1) print_filenames=true;
    while (argc > argument) strings_file(argv[argument++]);
    exit (EXIT_SUCCESS);
}

/* Print the strings.  Return true if ok, false if an error occurs.  */

strings_file (char *file)
{
  FILE *stream;

  if (file == NULL) {
/*    printf ("Pouzivam standardni vstup\n");      */
    print_strings (NULL, stdin, (file_ptr) 0, 0);
    return (EXIT_SUCCESS);
    }

/*  printf ("Otviram soubor %s\n", file); */

  stream = fopen (file, "r");
  if (stream == NULL)
    {
    fprintf (stderr, "%s: error opening \"%s\"\n",program_name,file);
    return (EXIT_FAILURE);
    }

  print_strings (file, stream, (file_ptr) 0, 0);

  fclose (stream);

  return (EXIT_SUCCESS);
}

/* Find the strings in file FILENAME, read from STREAM.
   Assume that STREAM is positioned so that the next byte read
   is at address ADDRESS in the file.
   Stop reading at address STOP_POINT in the file, if nonzero.

   If STREAM is NULL, do not read from it.
   Those characters come at address ADDRESS and the data in STREAM follow.  */

print_strings (char *filename, FILE *stream, file_ptr address, int stop_point)
{
  int bufsize;
  char *buf;

  bufsize = 100;
  buf = (char *) malloc (bufsize);

  while (1)
    {
      int i;
      int c;

      /* See if the next `string_min' chars are all graphic chars.  */
      
    tryline:
      if (stop_point && address >= stop_point)
        break;       /* skonci jakmile dosahne stop pointu */
        
      for (i = 0; i < string_min; i++)
        {
          if (stream == NULL)
            return(0);
          c = getc (stream);  /* Cte znak ze souboru */
          if (c < 0)
            return(0);

          address++;
          
          if (!isgraphic (c))
            /* Found a non-graphic.  Try again starting with next char.  */
            goto tryline;

          buf[i] = c;  /* Ulozi platny znak do bufferu */
        }

   /* Prvni cyklus nalezl prvnich 'string_min' po sobe jdoucich ascii znaku */

      /* We found a run of `string_min' graphic characters.
         Now see if it is terminated with a NUL byte or a newline.   */

      while (1)
        {
          if (i == bufsize)
            {
                             /* buffer se naplnil, zdvojnasobime ho */
              bufsize *= 2;
              buf = (char *) realloc (buf, bufsize);
            }
          if (stream == NULL)
            return(0);
          c = getc (stream);     /* cte znak ze souboru */
          if (c < 0)
            return(0);
 
         address++;

          if (c == '\0' || c == '\n')
            break;              /* It is; print this string.  */
          if (!isgraphic (c))
            goto tryline;       /* It isn't; give up on this string.  */

          buf[i++] = c;       /* The string continues; store it all.  */
        }

      /* If we get here, the string is all graphics and properly terminated,
         so print it.  It is all in `buf' and `i' is its length.  */

      buf[i] = '\0';      /* Zakonci string v bufferu */

      if (print_filenames)
        printf ("%s: ", filename);

      if (print_addresses)
        switch (address_radix)
          {
          case 8:
            printf ("%5lo ", (unsigned long) (address - i - 1));
            break;

          case 10:
            printf ("%5ld ", (long) (address - i - 1));
            break;

          case 16:
            printf ("%5lx ", (unsigned long) (address - i - 1));
            break;
          }

        for (i = 0; (c = buf[i]) != '\0'; i++)
        switch (c)
          {
          case '\n':
            printf ("\\n");
            break;
          case '\t':
            printf ("\\t");
            break;
          case '\f':
            printf ("\\f");
            break;
          case '\b':
            printf ("\\b");
            break;
          case '\r':
            printf ("\\r");
            break;
          default:
            putchar (c);    /* Tiskne string z bufferu */
          }
      putchar ('\n');
    }
}

/* ########################################### */

usage (FILE *stream, int status)
{
  fprintf (stream, "Usage: %s [-t d/o/x] [-n number] [file] ...\n", program_name);
  exit (status);
}

isascii_my (int c)
{
           /* spodni mala a horni velka pismena */
  if ((c>=32 && c<=127) || (c>=193 && c<=218))
    return (true);
  return (false);
}
